use crate::read;
use itertools::Itertools;

use std::{
    collections::*,
    hash::{Hash, Hasher},
};
#[derive(Debug, Clone, Eq)]
pub struct Coord {
    pub x: i32,
    pub y: usize,
    pub is_rock: bool,
}

impl Hash for Coord {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.x.hash(state);
        self.y.hash(state);
    }
}

impl PartialEq for Coord {
    fn eq(&self, other: &Self) -> bool {
        self.x == other.x && self.y == other.y
    }
}

impl Coord {
    pub fn down_left(&self) -> Self {
        Coord {
            x: self.x - 1,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
    pub fn down_right(&self) -> Self {
        Coord {
            x: self.x + 1,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
    pub fn down(&self) -> Self {
        Coord {
            x: self.x,
            y: self.y + 1,
            is_rock: self.is_rock,
        }
    }
}

pub fn run() {
    let input = read("14");
    let lines = input.lines();
    let mut rocks: HashSet<Coord> = lines
        .flat_map(|line| {
            line.split(" -> ")
                .map(|c| {
                    let nums = c.split_once(',').unwrap();
                    Coord {
                        x: nums.0.parse().unwrap(),
                        y: nums.1.parse().unwrap(),
                        is_rock: true,
                    }
                })
                .tuple_windows()
                .flat_map(|(s, e)| {
                    let mut path = vec![];
                    for nx in s.x..=e.x {
                        for ny in s.y..=e.y {
                            path.push(Coord {
                                x: nx,
                                y: ny,
                                is_rock: true,
                            })
                        }
                    }
                    for nx in e.x..=s.x {
                        for ny in e.y..=s.y {
                            path.push(Coord {
                                x: nx,
                                y: ny,
                                is_rock: true,
                            })
                        }
                    }
                    path
                })
                .collect_vec()
        })
        .collect();

    // Hide cursor
    print!("\x1B[?25l");
    #[cfg(feature = "part1")]
    {
        #[cfg(feature = "visualize")]
        print!("\x1B[2J\x1B[1;1H");

        let _count = p1(&mut rocks.clone());

        #[cfg(not(feature = "visualize"))]
        println!("Day 14, Part 01: {:?}", _count);
    }

    #[cfg(feature = "part2")]
    {
        let floor = rocks.iter().map(|r| r.y).max().unwrap() + 2;
        (-1000..=1000)
            .map(|x| Coord {
                x,
                y: floor,
                is_rock: true,
            })
            .for_each(|coord| {
                rocks.insert(coord);
            });
        // Clear screen
        #[cfg(feature = "visualize")]
        print!("\x1B[2J\x1B[1;1H");

        let _count = p2(&mut rocks);

        #[cfg(not(feature = "visualize"))]
        println!("Day 14, Part 02: {:?}", _count);
    }
    // Restore cursor
    print!("\x1B[?25h");
}

fn p1(rocks: &mut HashSet<Coord>) -> usize {
    let bottom = rocks.iter().map(|r| r.y).max().unwrap();
    let sand_start = Coord {
        x: 500,
        y: 0,
        is_rock: false,
    };
    let mut count_p1 = 0;

    loop {
        let mut sand = sand_start.clone();

        loop {
            if sand.y > bottom {
                #[cfg(feature = "visualize")]
                print(rocks, &sand, 1, count_p1);
                return count_p1;
            }
            let d = sand.down();
            let dl = sand.down_left();
            let dr = sand.down_right();

            if rocks.contains(&d) {
                if rocks.contains(&dl) {
                    if rocks.contains(&dr) {
                        count_p1 += 1;
                        rocks.insert(sand.clone());
                        #[cfg(feature = "visualize")]
                        print(rocks, &sand, 1, count_p1);
                        break;
                    }
                    sand = dr;
                } else {
                    sand = dl;
                }
            } else {
                sand = d;
            }
        }
    }
}

fn p2(rocks: &mut HashSet<Coord>) -> usize {
    let sand_start = Coord {
        x: 500,
        y: 0,
        is_rock: false,
    };
    let mut count_p2 = 0;

    loop {
        let mut sand = sand_start.clone();
        if rocks.contains(&sand) {
            #[cfg(feature = "visualize")]
            print(rocks, &sand, 2, count_p2);
            return count_p2;
        }

        loop {
            let d = sand.down();
            let dl = sand.down_left();
            let dr = sand.down_right();

            if rocks.contains(&d) {
                if rocks.contains(&dl) {
                    if rocks.contains(&dr) {
                        count_p2 += 1;
                        rocks.insert(sand.clone());
                        #[cfg(feature = "visualize")]
                        print(rocks, &sand, 2, count_p2);
                        break;
                    }
                    sand = dr;
                } else {
                    sand = dl;
                }
            } else {
                sand = d;
            }
        }
    }
}

#[cfg(feature = "visualize")]
fn print(rocks: &HashSet<Coord>, sand: &Coord, part: i32, count: usize) {
    let (min_y, min_x, max_y, max_x) = (10, 442, 182, 514);

    if sand.x < min_x || sand.x > max_x || sand.y < min_y || sand.y > max_y {
        return; // Don't render if not displayed
    }

    let status = format!("Part: {}, Count: {}", part, count);
    let bounds = " ".repeat((max_y as usize) - (min_y as usize) - status.len());
    let mut out = format!("{}{}\n", status, bounds);

    for x in (min_x..=max_x).rev() {
        for y in min_y..=max_y {
            let testing = &Coord {
                x,
                y,
                is_rock: false,
            };
            let c = match (sand == testing, rocks.get(testing)) {
                (true, _) => "\x1B[31mo\x1B[0m",
                (_, None) => " ",
                (_, Some(coord)) if coord.is_rock => "#",
                _ => "\x1B[33mo\x1B[0m",
            };
            out += c;
        }
        out += "\n";
    }
    out += &bounds;
    // Clear terminal, position cursor at (1,1)
    print!("\x1B[1;1H{}", out);
    let ms = if part == 1 { 34 } else { 8 }; // Go faster in part 2, takes too long otherwise
    std::thread::sleep(std::time::Duration::from_millis(ms));
}
